/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2013, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/

(function() {
dojo.provide("com.ibm.rdm.web.comparevvc.vvccompareutils");

dojo.require("com.ibm.rdm.web.comparevvc.VVCCompareConstants");
dojo.require("com.ibm.rdm.web.common.MessageHandler");

var VVCCompareConstants = com.ibm.rdm.web.comparevvc.VVCCompareConstants;
var MessageHandler = com.ibm.rdm.web.common.MessageHandler;

com.ibm.rdm.web.comparevvc.vvccompareutils = {
	
	
	isStructureConcept: function(/*string*/concept) {
		// Summary:
		//   Returns true if the supplied concept URI represents
		//   a module structure.
		//
		// concept:
		//   A URI string.
		
		var lastPathComponent = concept.substring(concept.lastIndexOf('/') + 1);
		return lastPathComponent.indexOf(VVCCompareConstants.STRUCTURE_SLUG_PREFIX) === 0;
	},
	
	getModuleURIForStructure: function(moduleStructureURI) {
		// summary:
		//		Given the URI of a module structure, return the URI of the corresponding module.
		// returns:
		//		A moduleURI, or a falsey value if the supplied URI is not that of a module structure.
		
		// Remove the prefix from the slug part of the structure URI to get the module URI  
		var structureSlugStart = moduleStructureURI.lastIndexOf('/') + 1,
			slugPrefix = VVCCompareConstants.STRUCTURE_SLUG_PREFIX,
			moduleSlugStart = structureSlugStart + slugPrefix.length;
		// no need to protect against moduleSlugStart being off the end of the string because slice() tolerates
		// index off the end.
		if (moduleStructureURI.slice(structureSlugStart, moduleSlugStart) === slugPrefix) {
			return moduleStructureURI.slice(0, structureSlugStart) + moduleStructureURI.slice(moduleSlugStart);
		}
		// return undefined
	},
	
	getStructureURIForModule: function(/*String*/moduleURI) {
		// Summary:
		//   Given a string URI of the form https://myserver:9443/resources/_7Wkfe345_8mcv
		//   which we assume refers to a module, generate the URi for the structure which
		//   involves sticking an S_ on the front of the last path component.
		//   so given the string above would return :
		//   https://myserver:9443/resources/S__7Wkfe345_8mcv
		//   
		if (!moduleURI) {
			throw "Attempt to build a structure URI without passing a moduleURI";
		}
		
		var lastIndexOfFS = moduleURI.lastIndexOf("/"),
	    	structureURI = VVCCompareConstants.STRUCTURE_SLUG_PREFIX + moduleURI.substring(lastIndexOfFS + 1),
	    	firstBit = moduleURI.substring(0, lastIndexOfFS + 1);
	    
	    return firstBit + structureURI;
	},
	
	isEmptyObject: function(obj) {
		for (var attr in obj) {
			return false;
		}
		
		return true;
	},
	
	addDeltaType: function(interimResultEntry, deltaType) {
		// summary:
		//		Helper for adding a deltaType to an item.
		// interimResultEntry:
		//		An interimResultEntry, in the format:
		//		{'concept': concept uri, 'deltaTypes': {'DELTA_TYPE_TARGET_ADD':true, ...}}
		// deltaType:
		//		One of the deltaType string values defined in CompareConstants,
		//		an array of these string values,
		//		or an object with keys of these string values and their corresponding
		//		entries set to true.
		
		if (!deltaType || !interimResultEntry) {
			return;
		}
		var deltaTypes = interimResultEntry.deltaTypes;
		if (!deltaTypes) {
			interimResultEntry.deltaTypes = deltaTypes = {};
		}
		
		if (typeof deltaType === 'string') {
			// The delta to add is a plain string value
			deltaTypes[deltaType] = true;
		}
		else if (dojo.isArrayLike(deltaType)) {
			// deltaType is an array
			// Grab each item and put it in the deltaTypes object
			dojo.forEach(deltaType, function(delta) {
				deltaTypes[delta] = true;
			});
		}
		else if (dojo.isObject(deltaType)) {
			// deltaType is an object
			dojo.mixin(deltaTypes, deltaType);
		}
	},
	
	hasDeltaType: function(interimResultEntry, deltaType) {
		// summary:
		//		Helper for determining whether an item has a given deltaType
		// interimResultEntry:
		//		An interimResultEntry, in the format:
		//		{'concept': concept uri, 'deltaTypes': {'DELTA_TYPE_TARGET_ADD':true, ...}}
		// deltaType:
		//		One of the deltaType string values defined in CompareConstants,
		//		an array of these string values,
		//		or an object with keys of these string values and their corresponding
		//		entries set to true.
		// returns:
		//		true if the interimResultEntry's deltaTypes has at least one element from
		//		deltaType in it.  Falsey value (undefined) if the interimResultEntry does 
		//		not have the at least one of the given deltaType entries.
		
		if (!deltaType || !interimResultEntry || !interimResultEntry.deltaTypes) {
			return; // falsey
		}
		
		var deltaTypes = interimResultEntry.deltaTypes;
		
		if (typeof deltaType === 'string') {
			// The delta to check is a plain string value
			return deltaTypes[deltaType];
		} else if (dojo.isArrayLike(deltaType)) {
			// The deltaType is an array of deltas
			for (var i = 0; i < deltaType.length; ++i) {
				// Just need one instance to return true
				if (deltaTypes[deltaType[i]]) {
					return true;
				}
			}
		} else if (dojo.isObject(deltaType)) {
			// Our deltaType to check is an object
			// Grab each item and check if they're ALL in our entries.
			// deltaType must be a subset of deltaTypes.
			
			for (var delta in deltaType) {
				// Just need one item in deltaType which is also in deltaTypes
				// to return successfully
				if (deltaTypes[delta]) {
					return true;
				}
			}
		}
		// return falsey
	},
	
	hasAllDeltaTypes: function(interimResultEntry, queryDeltaTypes) {
		// summary:
		//		Helper for determining whether an item has all 
		// interimResultEntry:
		//		An interimResultEntry, in the format:
		//		{'concept': concept uri, 'deltaTypes': {'DELTA_TYPE_TARGET_ADD':true, ...}}
		// queryDeltaTypes:
		//		An array of delta type string values, or an object with keys of these string
		//		values and their corresponding entries set to true.
		// returns:
		//		true if the interimResultEntry has all the queryDeltaTypes set on it.  Falsey value
		//		(undefined) if the interimResultEntry does not have all of the given queryDeltaTypes.
		
		if (!queryDeltaTypes || !interimResultEntry || !interimResultEntry.deltaTypes) {
			return;
		}
		
		var refDeltaTypes = interimResultEntry.deltaTypes;
		
		if (dojo.isArrayLike(queryDeltaTypes)) {
		
			// If the length of our filtered array isn't the same as the original,
			// something is missing and hence there is at least one value in 
			// queryDeltaTypes that isn't in refDeltaTypes
			
			return dojo.filter(queryDeltaTypes, function(delta) {
				return refDeltaTypes[delta];
			}).length === queryDeltaTypes.length;
		}
		// Our queryDeltaTypes to check is an object
		else if (dojo.isObject(queryDeltaTypes)) {
			// Grab each item and check if they're ALL in our entries.
			// queryDeltaTypes must be a subset of refDeltaTypes.
			
			for (var delta in queryDeltaTypes) {
				// Just need one item in queryDeltaTypes that isn't in refDeltaTypes
				// to return false
				if (!refDeltaTypes[delta]) {
					return false;
				}
			}
			return true;
		}
	},
	
	hasOnlyDeltaType: function(interimResultEntry, queryDeltaType) {
		// Summary:
		//  Helper to determine if an entry has only the given query delta type.
		return interimResultEntry.deltaTypes 
				&& Object.getOwnPropertyNames(interimResultEntry.deltaTypes).length === 1
				&& this.hasDeltaType(interimResultEntry, queryDeltaType);
	},
	
	getChangeEntriesByResourceType: function(/*Object*/entries, /*String*/resourceType) {
        // Summary:
        //
        //    Filters out changes of a particular type, e.g. changes
        //    to Object Types or Link Types as indicated by the argument
        //    resourceType, returns an array of changes of the supplied type 
        //    or an empty array if no matches are found.
        //
        //    e.g.
        //
        //    VVCCompareUtils.getChangeEntriesByResourceType(interimResultEntries, ResourceType.OBJECT_TYPE)
        var arrayOfTypeEntries = [],
            change;
        for (var uri in entries) {
            change = entries[uri];
            if (change.type === resourceType) {
                arrayOfTypeEntries.push(change);
            }
        }
        
        return arrayOfTypeEntries;
   },
   
   checkForLockedArtifactsError: function(error) {
   		// Summary:
   		// 		Checks to see if an XHR error is a result
   		//		of a 412 response, if it is we show the
   		//		message informing the user that some artifacts
   		//		they're trying to deliver are locked in the target.
   		//
   		//		Only used by code that uses the InterimResultsClient
   		//
   		// error:
   		//		XHR error response.
   		//
    	if (error.status === 412) {
			MessageHandler.showError(7739);
		} else { 
        	MessageHandler.showMessageForError(error);
        	return error;
        }
   }
};

})();
