/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2014, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
/*
Calculates and displays module structures for a module merge.  Two modules are shown.  The left
one is the source module, with changes highlighted on it.  The middle one is the merged module,
which is calculated from the targt module, with chages made to the source module merged into it.
The changes to be merged in are indicated by 'resource deltas' which come from the interim
result from VVC.  There may be additional differences between the source and target modules, either
because there are changes to the source that are not being delivered, or because of changes made
to the target.

In the code, the merged module representation is referred to as the 'middle'.  This is historical,
because at one time this widget displayed three module representations (source, merged, target).  
 
TODO:
TREE REPRESENTATION

Each displayed module grid is backed by a tree built of nodes that describe the tree to be displayed.
The tree contains references to rows mostly from one view, but it can contain references to rows from
other views (for example when some rows have been deleted from the view but we want to display them
as deleted, rather than just showing a placeholder).

The nodes in the tree can be modified as the user interact with the UI.

{
	content:
		rowData instance from a view.  The rowData itself must not be modified as it may
		be shared with other trees.
	children:
		An array of child nodes.  Empty array indicates that there are no children.  May be
		undefined in the case of one side of a move pair.
	structureNode:
		The structure node from whichever structure the node originated.  In a move pair, both
		tree nodes have a structureNode set, and they will be different structureNodes - one from
		each structure involved in the move.
	changeType:
		One of the constants from CompareConstants, indicating a change relating to the node.
	isForeign:
		For an added node, indicates that content has been created from content of another tree.
	movePartner:
		Reference to another node in the tree.  These nodes represent the alternative
		places that the content and children can live, depending on whether a move is
		accepted or not by the user.
	isForeignLocation:
		In a move pair, indicates the node that represents the merged in position of
		the moved node.
		Can be used in conjunction with presence/absence of children field to determine whether
		the move is considered 'actioned' or not.
	excluded: 
		True to indicate that the node should be considered absent from the merged tree.
		This allows the merged tree to contain nodes that are displayed as placeholders for
		added or removed nodes.  Changing the excluded flag allows the add or delete
		to effectively be overridden for a single node.  Only applies to nodes that have
		a changeType of added or removed.
		A node that is excluded can have child nodes that are not excluded.  In such a case
		the children are considered to be lifted up the hierarchy to the position of the
		excluded parent.  Thus we can have a coherent merged tree even though some nodes
		that are serving as parents in the hierarchy are excluded, but their children are
		included.
}

For the left and right trees, only content, structureNode, children, changeType and isForeign are used.
isForeign is only applicable for removed rows.  For a removed row the content comes
from the view on the the other tree, if available.  If there is no corresponding row on the
other tree, a fake content object is used.

Most of the tree complexity is used in the middle tree, representing the merged structure.
The user is expected to do operations on the middle tree, which can affect the state of the
tree, and hence the state of the merged module.

For a move there are two nodes in the tree - one at each of the two positions that the content
and children could be located (depending on whether the user chooses to action the move or not).
The two nodes reference each other through their movePartner references.  One has isForeignLocation
set, to indicate that it is the 'merged in' position.
Only one of the two nodes has content and children.  As the move is accepted or rejected by the user,
the content and children are swapped to the other partner in the pair - so the presence/absence of
content and children indicates the selected position.

In the middle tree, a node that is deleted is fully represented in the tree, but it is generally
rendered as a placeholder only.

DISPLAY OF TREES

Based on the information in the nodes the tree nodes are displayed in various ways.

In the middle tree:
Deletes that come from the merge target are shown as placeholders, and are represented by a simple
placeholder node.  Adjacent placeholders are merged into one (containing the IDs of all the merged in
nodes).  The user cannot 'reject' a delete that is already in the target, so there is no need
to display separate placeholder for each deleted row.
Deletes that come from the merge source are included as full nodes, but are displayed with styling
that makes them look like placeholders.  This gives a place for the user to interact with each row.a

For moves, one of the pair of nodes is displayed as a full node, and the other is displayed as
a placeholder.  This allows the user to see the alternative position of each move.

*/

(function() {
dojo.provide("com.ibm.rdm.web.merge.ModuleStructureMergeWidget");

dojo.require("dijit._Widget");
dojo.require("dijit._Templated");
dojo.require("dojox.gfx");

// Use absolute minimum of requires, so that we can get run this widget in stand-alone test fixture.
dojo.require("com.ibm.rdm.web.compare.CompareConstants");
dojo.require("com.ibm.rdm.web.compare.moduleCompareUtils");
dojo.require("com.ibm.rdm.web.compare._ModuleCompareWidgetBase");
dojo.require("com.ibm.rdm.web.comparevvc.vvccompareutils");
dojo.require("com.ibm.rdm.web.merge.ModuleMergeTreeSerializer");
dojo.require("com.ibm.rdm.web.util.domutils");
dojo.require("com.ibm.rdm.web.common.OverlayBorder");
dojo.require("com.ibm.rdm.web.model.types.RRMTypes");

dojo.require("com.ibm.rdm.web.merge.ModuleStructureMergeDataProvider");
dojo.require("com.ibm.rdm.web.merge.ModuleStructureAutoMerger");
dojo.require("com.ibm.rdm.web.model.resources.modules.SectionNumberingStrategy");

// Use the same message file as the similar compare widget
dojo.requireLocalization("com.ibm.rdm.web.compare", "ModuleStructureCompareWidget");
dojo.requireLocalization("com.ibm.rdm.web.merge", "ModuleStructureMergeWidget");

var CompareConstants = com.ibm.rdm.web.compare.CompareConstants,
	ModuleCompareUtils = com.ibm.rdm.web.compare.moduleCompareUtils,
	VVCCompareUtils = com.ibm.rdm.web.comparevvc.vvccompareutils,
	DomUtils = com.ibm.rdm.web.util.domutils,
	ModuleMergeTreeSerializer = com.ibm.rdm.web.merge.ModuleMergeTreeSerializer,
	sectionNumberer = com.ibm.rdm.web.model.resources.modules.SectionNumberingStrategy.getDefault(),
	RRMTypes = com.ibm.rdm.web.model.types.RRMTypes,
	_ModuleCompareWidgetBase = com.ibm.rdm.web.compare._ModuleCompareWidgetBase,
	ModuleStructureAutoMerger = com.ibm.rdm.web.merge.ModuleStructureAutoMerger,
	mscMessages = dojo.i18n.getLocalization("com.ibm.rdm.web.compare", "ModuleStructureCompareWidget");


var ROW_URI_ATTR = "data-rdm-uri",
	ROW_CONTENT_WIDTH_ADJUST_ATTR = "data-rdm-width-adjust",
	CONTENT_COL_HEADING_TAG_ATTR = "data-rdm-heading-tag",
	CONNECTION_LINE_COLOR = "#999999",
	CONNECTION_LINE_COLOR_ADDED = CONNECTION_LINE_COLOR, // d6e6ee
	CONNECTION_LINE_COLOR_MOVED = "#80a2be",
	CONNECTION_MOVE_PAIR_LINE_COLOR = "#d6e6ee",
	CONNECTION_MOVE_PAIR_LINE_COLOR_BOLD = "#57bcee",
	CONNECTION_SURFACE_WIDTH = 80,
	CONNECTION_SURFACE_DEFAULT_HEIGHT = 600,
	CONNECTION_CURVE_CONTROL_LEFT = 40,
	CONNECTION_CURVE_CONTROL_RIGHT = CONNECTION_SURFACE_WIDTH - CONNECTION_CURVE_CONTROL_LEFT,
	CONNECTION_MOVE_PAIR_CURVE_CONTROL_LEFT = 40,
	CONNECTION_MOVE_PAIR_CURVE_CONTROL_RIGHT = CONNECTION_SURFACE_WIDTH - CONNECTION_MOVE_PAIR_CURVE_CONTROL_LEFT,
	CONNECTION_ARROW_WIDTH = 8,
	CONNECTION_ARROW_HEIGHT = 4,
	CONNECTION_ARROW_CURVE_END_LEFT = 10,
	CONNECTION_ARROW_CURVE_END_RIGHT = CONNECTION_SURFACE_WIDTH - CONNECTION_ARROW_CURVE_END_LEFT,
	CONNECTION_OFFSET = 8;


function createOrAddRemovePlaceholder(rowData, rowId, previousRow, contentParent, moveBlocksToClose) {
	// summary:
	//		Makes a remove placeholder, or adds a new entry to the remove placeholder
	//		of the previous row.
	// rowData:
	//		View data for the row.
	// rowId:
	//		ID attribute for the row (not row URI - this is the unique HTML id to use).
	// previousRow:
	//		DOM Node for the previous row, which may itself be a placeholder.
	// contentParent:
	//		DOM Node to add the placeholder to.
	// moveBlocksToClose:
	//		The number of move blocks to close on the row.  This can update classes on the existing previous row
	//		or on the current row being added.
	// returns
	//		The created placeholder if one was created.  If the entry was added to an existing placeholder in
	//		the previous row, undefined is returned.
	var createdPlaceholder,		// is set only if a new placeholder is created.
		placeholder,			// is set to the existing placeholder or the newly created one
		placeholderEntry,
		nestingNode;
	if (!dojo.hasClass(previousRow, 'removed') ||
		!(placeholder = dojo.query('.placeholder', previousRow)[0])) {		// assignment - single equals is deliberate
		if (!placeholder) {
			// need to create a new placeholder
			placeholder = createdPlaceholder = dojo.create('div', {'class': 'placeholder'}, contentParent);
		}		
	}
	// Create the entry within the placeholder.
	placeholderEntry = dojo.create('div', {'class': 'placeholderEntry', id: rowId}, placeholder);
	dojo.attr(placeholderEntry, ROW_URI_ATTR, rowData.getResourceURI());

	// close move blocks if necessary
	nestingNode = placeholder;
	while (moveBlocksToClose-- && nestingNode) {
		nestingNode = DomUtils.findAncestorByClass(nestingNode, "moved");
		if (nestingNode) {
			dojo.addClass(nestingNode, "blockEnd");
		}
	}
	
	return createdPlaceholder;
}


dojo.declare("com.ibm.rdm.web.merge.ModuleStructureMergeWidget", [dijit._Widget, dijit._Templated, _ModuleCompareWidgetBase], {
	templatePath: dojo.moduleUrl("com.ibm.rdm.web", "merge/templates/ModuleStructureMergeWidget.html"),
	widgetsInTemplate: true,
	// Option that may be set on construction.  If true, the widget will show button to
	// switch moves.  Left in for debug only.
	showMoveActionButtons: false,
	
	// connect points
	_leftGrid: null,
	_middleGrid: null,
	
	// The surface for drawing connections
	_leftConnectionSurface: null,
	
	// The current height of the connection surface.
	_connectionSurfaceHeight: CONNECTION_SURFACE_DEFAULT_HEIGHT,

	// Array of shapes that are used to connect a highlighted move connector.
	_moveHighlightConnections: null,

	// Array of all connections that are on display.	
	_displayedConnections: null,
	
	// object that provides a load function which gets called during show
	// to fetch the data required to render the trees.
	_mergeDataProvider: null, 
	messages: dojo.i18n.getLocalization("com.ibm.rdm.web.merge", "ModuleStructureMergeWidget"),
	// The root nodes of the data trees that back the grid
	_leftTree: null,
	_middleTree: null,
	
	// A helper class that is injected during show().  Used to fetch the detailed artifact
	// content in a form that can be displayed.  Injected so that an alternative test version
	// of it can be provided for test purposes. 
	_artifactContentLoader: null,
	
	
	// Some prefixes used for building DOM node IDs.
	_LEFT_ID_PREFIX: null,
	_RIGHT_ID_PREFIX: null,
	_MIDDLE_ID_PREFIX: null,
	_MIDDLE_INCOMING_ID_PREFIX: null,
	
	postCreate: function() {
		this._leftConnectionSurface = dojox.gfx.createSurface(this._leftConnectionSurfaceNode, CONNECTION_SURFACE_WIDTH, this._connectionSurfaceHeight);
		this._displayedConnections = [];
		this._moveHighlightConnections = [];
		// Set up some prefixes to be used for building unique DOM node IDs.
		this._LEFT_ID_PREFIX = this.id + '_L_';
		this._RIGHT_ID_PREFIX = this.id + '_R_';
		this._MIDDLE_ID_PREFIX = this.id + '_M_';
		this._MIDDLE_INCOMING_ID_PREFIX = this.id + '_MI_';
		
		this._mergeDataProvider = new com.ibm.rdm.web.merge.ModuleStructureMergeDataProvider();
		
		// Connect up some event handlers
		this.connect(this._leftGrid, "onclick", dojo.hitch(this, this._onGridClick, this._leftGrid));
		this.connect(this._middleGrid, "onclick", dojo.hitch(this, this._onGridClick, this._middleGrid));
		
		// alias the _middleGrid with _rightGrid so we can use some common
		// functions in the base class
		this._rightGrid = this._middleGrid;
		this._rightGridScroller = this._middleGridScroller;
	},
	
	resize: function(newWidth, newHeight) {
		this.width = newWidth < 300 ? 300 : newWidth;
		this.height = newHeight < 10 ? 10 : newHeight;
		
		if (this._scale) {
			this.width = this.width / this._scale;
			this.height = this.height / this._scale;
			
			//CONNECTION_SURFACE_WIDTH = CONNECTION_SURFACE_WIDTH / this._scale;
		}
		
		this._resizeImpl();
	},
	
	scale: function(by) {
		if (this._scale) {
			this.width = this.width * this._scale;
			this.height = this.height * this._scale;
			//CONNECTION_SURFACE_WIDTH = CONNECTION_SURFACE_WIDTH * this._scale;
		}
		this._scale = by;
		dojo.style(this.domNode, "transform", "scale("+by+")");
		this.resize(this.width, this.height);
		
		this._showAnnotations();
	},
	
	_resizeImpl: function() {
		var totalWidthAvailable = this.width,
			leftWidth = Math.floor((totalWidthAvailable - CONNECTION_SURFACE_WIDTH) / 2),
			middleWidth = this.width - leftWidth - CONNECTION_SURFACE_WIDTH, // not necessarily exactly the same as left width, to allow for rounding, else
																			 // its right-hand edge can jitter away from the connectors.
			totalHeightAvailable = this.height,
			gridHeight = totalHeightAvailable - 1, // allow for 1px top border
			gridHeightStr = gridHeight + "px";
			
		dojo.style(this.domNode, {
			width: totalWidthAvailable+"px",
			height: totalHeightAvailable+"px"
		});
		
		dojo.style(this._gridArea, {
			width: totalWidthAvailable+"px",
			height: gridHeightStr
		});

		dojo.style(this._leftSizeWrapper, {
			width: leftWidth+"px",
			height: gridHeightStr
		});
		
		dojo.style(this._leftGrid, {
			height: gridHeightStr
		});

		dojo.style(this._middleSizeWrapper, {
			width: middleWidth+"px",
			height: gridHeightStr
		});
		
		dojo.style(this._middleGrid, {
			height: gridHeightStr
		});
		
		//this._resizeWideContentNodes();
		
		dojo.style(this._leftConnectionSurfaceNode, {
			width: CONNECTION_SURFACE_WIDTH,
			left: leftWidth+"px"
		});
		this._leftConnectionSurface.setDimensions(CONNECTION_SURFACE_WIDTH, gridHeight);
		this._connectionSurfaceHeight = gridHeight;
		this._showConnections();
		this._redrawSelectionOverlays();
		this.onColumnResize(leftWidth, middleWidth, leftWidth + CONNECTION_SURFACE_WIDTH);
	},
	
	onColumnResize: function(leftWidth, rightWidth, rightPosition) {
		
	},
	
	clear: function() {
		// summary:
		//		Clears the widget's display
		dojo.empty(this._leftGrid); 
		dojo.empty(this._middleGrid);
		this._clearAnnotations();
		dojo.removeClass(this.domNode, "noChanges");
		this._redrawSelectionOverlays();
		delete this._selectedMovePair;
		delete this._rightSelection;
		this._rendered = false;
		delete this._leftRows;
	},

	show: function(params) {
		// Summary:
		//  Called by containing widget (TabbedArtifactWidget) to show this widget.
		//
		//  params:
		//		An object with the following properties :
		//		sourceOnLeft,
		//		mergeStateStore, 
		//		targetResource,
		//		targetType,
		//		targetConfiguration,
		//		targetConfigurationName,
		//		sourceResource,
		//		sourceType,
		//		sourceConfiguration,
		//		sourceConfigurationName,
		//		primaryResource,
		//		primaryType,
		//		primaryAttributeDefinitions,
		//		attributeDifferences,
		//		publicTagDifferences,
		//		privateTagDifferences,
		//		linksDifferences,
		//		displayType,
		//		resourceDeltas,
		//		operation

		this._artifactContentLoader = params.artifactContentLoader;
		this.onColumnResize = params.onColumnResize;
		this._compareContext = {
			targetConfiguration: params.targetConfiguration,
			sourceConfiguration: params.sourceConfiguration
		};
		
		this.clear();
		
		// Find the concept URI for the Module Structure in the resourceDeltas... 
		
		var resourceDeltas = params.resourceDeltas,
			// this is the object we pass to the MergeDataProvider 
			// after we extract the correct parameters from the object
			// passed to this function
			providerArgs, 
			moduleURI = params.primaryResource.resourceURI,
			structureChangeEntry = resourceDeltas[VVCCompareUtils.getStructureURIForModule(moduleURI)],
			// two functions chained onto the providers load to...
			// show the grids
			showGrids = dojo.hitch(this, this._showGrids);
		this._resourceDeltas = resourceDeltas;
		if (structureChangeEntry) {
			return ModuleStructureAutoMerger.getMergedStructure(moduleURI)
				.then(function(mergedChanges) { return dojo.mixin(params, mergedChanges); })
				.then(showGrids);	
		} else {
			dojo.addClass(this.domNode, "noChanges");
		}
	},
	
	onMergeTreeChange: function() {
		// summary:
		//		Called whenever the merge tree changes.  The containing widget can use this to trigger a
		//		call to getMergeTreeRDF().
		//		Container can provide an alternative implementation or connect to this method.
	},

	getMergeTreeRDF: function() {
		// returns:
		//		 An RDF XML representation of the merged tree.
		var serializer = new ModuleMergeTreeSerializer(true /*pretty*/);
		return serializer.toRDFString(this._middleTree);
	},

	_showGrids: function(params) {
		// Display the three grid areas.
	
		var structureChanges = params.structureChanges,
			valueChanges = params.valueChanges;
		
		this._leftTree = structureChanges.leftTree.getRoot();
		this._middleTree = structureChanges.middleTree.getRoot();
		
		sectionNumberer.number(this._middleTree, com.ibm.rdm.web.merge.ModuleStructureMergeWidget.SECTION_NUMBER_MERGE_TREE_ARGS); 
		sectionNumberer.number(this._leftTree, com.ibm.rdm.web.merge.ModuleStructureMergeWidget.SECTION_NUMBER_LEFT_TREE_ARGS); 

		this._drawGrid(this._leftGrid, structureChanges.leftTree.getRoot(), false, valueChanges);
		this._drawGrid(this._middleGrid, structureChanges.middleTree.getRoot(), /*isMergedTree*/true, valueChanges);
		this.onMergeTreeChange();
		this._showConnections();
		this._connectImgLoadToAnnotationRedraw();
		this._showAnnotations();
		this._rendered = true;
	},
	
	_drawGrid: function(gridNode, viewTreeRoot, isMergedTree, valueChangeMap) {
		// summary:
		//		Draw the content for a grid.
		// gridNode:
		//		The node to which grid rows are to be added.
		// viewTreeRoot:
		//		The root of a tree to display in the grid.
		// idGenerator:
		//		A function that generates the ID for each row in the grid.  Takes the resourceURI as parameter.
		// isMergedTree:
		//		true to indicate that the grid being drawn is the merged grid (which has some slightly different display
		//		rules, and is the tree in which interaction takes place).
		

		var idGenerator = (gridNode === this._leftGrid) ? this._makeLeftNodeId : this._makeMiddleNodeId,
			i, rowData, rowURI, previousRow, removedBlockSize, addedBlockSize;
			
		idGenerator = dojo.hitch(this, idGenerator);
		isMergedTree = (gridNode === this._middleGrid);
		// Ensure the grid is clear to start with.
		dojo.empty(gridNode);
		if (gridNode === this._middleGrid && this._moveToggleConnectHandles) {
			this._moveToggleConnectHandles.forEach(function(handle){
				this.disconnect(handle);
			}, this);
		}
		this._moveToggleConnectHandles = [];
			
		var drawRow = dojo.hitch(this, function(rowTreeNode, depth, moveNestTypeStack, isMoveBlockStart, moveBlocksToClose, isMovePlaceholder, isForeignLocation) {
			// summary:
			//		Adds a single row to the grid.
			//		This method is called in order for each row required by the outer _drawGrid function.  It makes
			//		use of several variables scoped at the _drawGrid level to provide continuity between consecutive calls. 
			// depth:
			//		The depth of the row to be drawn.
			// rowTreeNode:
			//		The tree node for the row to be drawn.
			//console.log("drawRow drawing", rowTreeNode, depth, moveNestTypeStack, isMoveBlockStart, moveBlocksToClose);
			var rowData = rowTreeNode.content;
			
			if (!rowData) {
				return;
			}
			
			
			
			var	changeType = rowTreeNode.changeType,
				rowURI = rowData.getResourceURI(),
				rowId, rowClasses, row, contentParent, isRemovePlaceholder, createdPlaceholder,
				valueChanged = !!(valueChangeMap && valueChangeMap[rowURI]);
			
			if (valueChanged && isMergedTree) {
				var leftRow = valueChangeMap[rowURI];
				if (leftRow && (typeof leftRow === "object")) {
					rowData = leftRow;
				}
			}
			
			// Create the outline of the row
			rowClasses = ["row"];
			row = dojo.create("div");
			rowClasses.push((depth >= 10) ? "depth10plus" : "depth"+depth);
			contentParent = row;
			
			// Add move blocks, which may nest, and add/remove blocks.
			contentParent = this._addMovedRowDecorations(contentParent, moveNestTypeStack, isMoveBlockStart, moveBlocksToClose, rowClasses, rowTreeNode);
			
			switch (changeType) { // could be undefined, but that's OK
				case CompareConstants.STRUCTURE_DIFF_TYPE_ADD:
					// An add block
					if (isMergedTree && rowTreeNode.excluded) {
						// this is a tree node for an add that is not being delivered.  Don't
						// display it at all in the merged tree.
						return;
					}
					contentParent = dojo.create('div', {"class": 'block blockStart blockEnd added'}, contentParent);
					rowClasses.push('added');
					rowClasses.push('blockStart');
					if (rowTreeNode.isForeign) {
						dojo.addClass(contentParent, 'incoming');
						rowClasses.push('incoming');
						// For additions we normally show the content, unless the user has previously overridden this to
						// remove the row back from the merged display.  We will always show content on first display, but
						// on a redraw of this grid we need to maintain the changes that the user has already made.
						if (rowTreeNode.excluded) {
							dojo.addClass(contentParent, "hideContent");
						}
					}
					break;
				case CompareConstants.STRUCTURE_DIFF_TYPE_REMOVE:
					// A remove block
					if (isMergedTree) {
						if (rowTreeNode.isForeign) {
							// row removed from the 'to' tree, so display all consecutive removed rows as one placeholder.
							rowId = idGenerator(rowURI, false);
							createdPlaceholder = createOrAddRemovePlaceholder(rowData, rowId, previousRow, contentParent, moveBlocksToClose);
							if (!createdPlaceholder) {
								// placeholder entry was added to the previous row.  Don't add this row.
								return;
							}
							// new placeholder was created in the row
							rowClasses.push('blockStart');
							isRemovePlaceholder = true;
							contentParent = createdPlaceholder;
						} else {
							// row removed from the 'from' tree, so include the content in the tree so that it can be
							// shown/hidden, but initially hide the content, so it looks like a placeholder.
							contentParent = dojo.create('div', {"class": 'block blockStart blockEnd removed incoming'}, contentParent);
							// For deletions we normally hide the content, unless the user has previously overridden this to
							// bring the row back into the merged display.  We will always hideContent on first display, but
							// on a redraw of this grid we need to maintain the changes that the user has previously made.
							if (rowTreeNode.excluded) {
								dojo.addClass(contentParent, "hideContent");
							}
							
							rowClasses.push('blockStart');
							rowClasses.push('incoming');
						}
					} else {
						// A deleted row.
						if (rowTreeNode.excluded) {
							// The row should be excluded completely from display.  Don't add this row.
							return;
						}
						contentParent = dojo.create('div', {"class": 'block blockStart blockEnd removed'}, contentParent);
						rowClasses.push('blockStart');
					}
					rowClasses.push('removed');
					break;
				case CompareConstants.STRUCTURE_DIFF_TYPE_MOVE:
					// A move block - display it only if it is a placeholder - otherwise it is just decoration on rows, not a row in itself.
					if (isMovePlaceholder) {
						contentParent = dojo.create('div', {"class": 'block blockStart blockEnd moved'}, contentParent);
						if (isForeignLocation) {
							dojo.addClass(contentParent, 'incoming');
							rowClasses.push('incoming');
						}
						// Add the event handlers for hover and button to toggle the position of the row (i.e. accept/not accept the move)
						this._addMoveHoverControls(contentParent);
						rowClasses.push('moved');
						rowClasses.push('blockStart');
					}
					break;
			}			

			// Add the row content direct on the row, or the innermost block, unless we are drawing a placeholder, in which case just add
			// some classes for styling.
			if (isMovePlaceholder || isRemovePlaceholder) {
				dojo.addClass(contentParent, 'placeholder');
			} else {
				// normal (non-placeholder) row
				var allowShowDetails = isMergedTree || (changeType === CompareConstants.STRUCTURE_DIFF_TYPE_REMOVE);
				contentParent.appendChild(this._buildRowContent(rowData, depth, rowTreeNode.structureNode, allowShowDetails, valueChanged, isMergedTree));
			}
			
			// Add indication of changed value
			if (valueChanged) {
				rowClasses.push("valueChanged");
				var thisDelta = this._resourceDeltas[rowURI];
				if (VVCCompareUtils.hasOnlyDeltaType(thisDelta, CompareConstants.DELTA_TYPE_LINKS_CHANGE)) {
					rowClasses.push("linkEndpointOnly");
				}
			}
			
			// remove placeholders have the id set on a node within the placeholder, so avoid setting it on the row as well.
			// Rows that are the start of an incoming operation are given a slightly different ID, so that we do not have
			// duplicate IDs on move blocks and their corresponding placeholder rows.
			dojo.addClass(row, rowClasses);
			rowId = idGenerator(rowURI, dojo.hasClass(row, 'incoming'));
			if (!isRemovePlaceholder) {
				dojo.attr(row, 'id', rowId);
				dojo.attr(row, ROW_URI_ATTR, rowURI);
			}
	
			// suppress the dividing line immediately before the start of a block.
			if (previousRow && (dojo.hasClass(contentParent, 'blockStart') || isRemovePlaceholder)) {
				dojo.addClass(previousRow, "noDivider");
			}
			
			gridNode.appendChild(row);
			previousRow = row;
		});		// end of drawRow()
		
		var displayTree = dojo.hitch(this, function(parent, depth, moveNestTypeStack, moveBlocksToClose) {
			// summary:
			//		Displays a tree by recursively walking the nodes and their children.  The root node
			//		is not drawn.
			// parent:
			//		The node to be drawn.  All of its children, grandchildren etc. are drawn as well.
			// depth:
			//		The depth in the tree of the parent.  The root node of the tree is depth 0.
			// moveNestTypeStack:
			//		Move operations can be nested within each other. This value indicates the depth move
			//		nesting that exists around the parent (not including any move of the parent itself).  Entries in
			//		the stack are true to indicate that the move operation came from the source (left) side, and falsey
			//		to indicate that the move came from the target side of the merge.
			// moveBlocksToClose:
			//		The number of nested move blocks to close on this row.  Only one move block can be opened
			//		on a row, but multiple blocks can be closed on a single row - when the end of several nested
			//		move blocks coincide.
			 
			// Before displaying the row, see if it is the start of a move block.  If it is, it and all its children
			// need to be given an additional move nesting
			
			var hasChildren = parent.children && parent.children.length,
				moveBlockStart, currentRowMoveBlocksToClose;
			if (parent.changeType === CompareConstants.STRUCTURE_DIFF_TYPE_MOVE) {
				moveNestTypeStack.push(parent.isForeignLocation);
				moveBlockStart = true;
				if (!hasChildren) {
					// no children, so close this new move block when we draw the row.
					moveBlocksToClose++;
				}
			}
			
			currentRowMoveBlocksToClose = hasChildren ? 0 : moveBlocksToClose;
			
			// If the node is the placeholder end of a move partner pair, display it as a placeholder
			if (parent.movePartner && !parent.children) {
				// The empty end of a move pair.  Display it as a move placeholder.
				drawRow(parent.movePartner, depth, moveNestTypeStack, moveBlockStart, currentRowMoveBlocksToClose, true/*isMovePlaceholder*/, parent.isForeignLocation);
			} else {
				// A normal row
				// display the row - but only if it is not the root node.
				if (depth) {
					drawRow(parent, depth, moveNestTypeStack, moveBlockStart, currentRowMoveBlocksToClose, false/*isMovePlaceholder*/, parent.isForeignLocation);
				}
				if (moveBlockStart && hasChildren) {
					// has children, so close the move block on the deepest last child
					moveBlocksToClose++;
				}
							
				// prepare for recursion to display children
				var childDepth = depth + 1;
				
				// If we are on an add or remove that is hidden (i.e. not included in the merged module),
				// adjust the depth of children so that they are raised up the tree to account for the
				// absent parent node.
				if (parent.excluded) {
						// A node that is not included in the merged tree.  Pull its children up one so that it
						// appears at the same depth as the excluded parent.
						childDepth--;
				}
				
				// do the recursion for each child
				var lastChildIndex = parent.children.length-1;
				dojo.forEach(parent.children, function(child, index) {
					var childMoveBlocksToClose = 0;
					if (index === lastChildIndex) {
						// last child of this parent.
						childMoveBlocksToClose = moveBlocksToClose;
					}
					displayTree(child, childDepth, moveNestTypeStack, childMoveBlocksToClose);
				}, this);
			}
			// The row and any child rows have been displayed
			if (moveBlockStart) {
				// All children displayed, so if the row was a move, take it off the move stack, as we have now
				// displayed the end of the move block.
				moveNestTypeStack.pop();
			}
		});
		
		// Start off the recursion that draws the tree.
		displayTree(viewTreeRoot, 0, [], 0);
	},

	
	_buildRowContent: function(rowData, depth, structureTreeNode, allowShowDetails, valueChanged, isMergedTree) {
		// summary:
		//		Builds a DOM node to represent the content of a row.  This does not include
		//		the nesting wrappers for change outline boxes.

		// rowData:
		//		Partial Requirement to be displayed.
		// depth:
		//		The depth of the row in the tree hierarchy.
		// structureTreeNode:
		//		The structure node from which the row is derived.  Can be from any of the
		//		structure trees, depending on the origin of the row in the displayed tree.
		// allowShowDetails:
		//		truthy to indicate that a 'show content' operation should be permitted on the row.
		// valueChanged:
		//		truthy to indicate that the content has changed.
		// returns:
		//		A DOM node containing a display representation of the row.
		
		var contentWrapper = dojo.create("div", {"class": "contentWrapper"}),
			innerBox = dojo.create("div", {"class": "innerBox"}, contentWrapper),
			table = dojo.create('table', {"class": "rowCells"}, innerBox),
			tbody = dojo.create('tbody', null, table),
			row = dojo.create('tr', null, tbody),
			idCell = dojo.create('td', {"class": "idColumn"}, row),
			ptCell = dojo.create('td', {"class": "contentColumn"}, row),
			displayContent = dojo.create('div', {"class": "displayContent initialContent show"}, ptCell);

		idCell.appendChild(document.createTextNode(structureTreeNode.boundIdentifier));
		//idCell.appendChild(document.createTextNode("Book order "+structureTreeNode.bookOrder));
		//idCell.appendChild(document.createTextNode("isForeign "+structureTreeNode.isForeign));
		
		if (allowShowDetails) {
			// Add a link/button.  Exact function depends on whether the value in the row has changes or not.
			var selectorText = mscMessages.showContent,
				selectorClasses = "showContentLink";
			if (valueChanged && !rowData.getWrappedResourceMimeType()) {
				selectorText = mscMessages.showChanges;
				selectorClasses += " showChanges";
			}
			var selector = dojo.create('div', {"class":"contentSelector"});
			var link = dojo.create('a', {"class":selectorClasses,"tabIndex":"-1"}, selector);
			link.setAttribute('href', '#');
			link.appendChild(document.createTextNode(selectorText));
			displayContent.appendChild(selector);
		}
		
		
		if (rowData.getWrappedResourceMimeType()) {
			// It's a wrapped resource
			this._displayWrappedResource(displayContent, rowData, this._compareContext.sourceConfiguration);
		} else if (structureTreeNode.isHeading) {
			var headingTag = 'h' + (depth > 6 ? 6 : depth);
			var headingNode = dojo.create(headingTag, {"class":"textContent"});
			var sectionNumberElement = dojo.create('div', {"class": "sectionNumber"}, headingNode);
			if (isMergedTree) {
				sectionNumberElement.appendChild(document.createTextNode(structureTreeNode.section));
			} else {
				sectionNumberElement.appendChild(document.createTextNode(structureTreeNode.sectionNumberPreMerge));
			}
			var headingTextElement = dojo.create('div', {"class": "headingText"}, headingNode);
			headingTextElement.innerHTML = rowData.getAttributeValue(RRMTypes.WELL_KNOWN_SAME_AS_ATTRIBUTE_NAMES.PrimaryText);//rowData.getPrimaryText(); 
			displayContent.appendChild(headingNode);
			// Indent the heading text to right of the section number
			// Add the heading tag as an attribute to the contentColumn node so that
			// it can be found for re-display of the column content when we don't have
			// the full rowData available.
			dojo.attr(ptCell, CONTENT_COL_HEADING_TAG_ATTR, headingTag);
		} else {
			var ptWrapper = dojo.create('div', {"class":"textContent"}, displayContent);
			ptWrapper.innerHTML = rowData.getAttributeValue(RRMTypes.WELL_KNOWN_SAME_AS_ATTRIBUTE_NAMES.PrimaryText);//rowData.getPrimaryText() + " D=" + depth;
		}
	
		return contentWrapper;
	},
	
	_clearConnections: function() {
		// summary:
		//		Clear all connections from display
		dojo.forEach(this._displayedConnections, function(connection){
			connection.removeShape();
			//connection.destroy();
		});
		this._displayedConnections = [];

		this._clearMoveHighlightConnection();
	},
	
	_clearMoveHighlightConnection: function() {
		if (this._moveHighlightConnections) {
			this._moveHighlightConnections.forEach(function(shape){
				shape.removeShape();
			});
			this._moveHighlightConnections = [];
		}
	},

	
	_addMovedRowDecorations: function(contentParent, moveNestTypeStack, isMoveStart, moveCloseCount, rowClasses, rowTreeNode) {
		// If there is any nesting row should have move styling
		if (moveNestTypeStack.length) {
			rowClasses.push('moved');
		}
		// Make an array of nested nodes, and keep them in a stack so we can add start and end classes.
		var nestingNodes = [];
		for (var i = 0 ; i < moveNestTypeStack.length ; i++) {
			nestingNodes.push(contentParent = dojo.create('div', {"class": 'block moved'}, contentParent));
			if (moveNestTypeStack[i]) {
				dojo.addClass(contentParent, 'incoming');
			}
		}
		// add start move decoration if required.
		var lastIndex = nestingNodes.length-1;
		if (isMoveStart) {
			dojo.addClass(nestingNodes[lastIndex], 'blockStart');
			rowClasses.push('blockStart');
			if (moveNestTypeStack[lastIndex]) {
				rowClasses.push('incoming');
			}
			// Add a button to toggle the position of the row (i.e. accept/not accept the move), if the move is
			// part of a move pair - not just a display of a move made to the target.
			if (rowTreeNode.movePartner) {
				this._addMoveHoverControls(contentParent);
			}
		}
		for (var index = lastIndex ; moveCloseCount > 0 ; moveCloseCount--, index--) {
			dojo.addClass(nestingNodes[index], 'blockEnd');
		}
		return contentParent;
	},

	_showConnections: function() {
		this._clearConnections();
		
		this._showConnectionsFromGrid(this._leftGrid, this._leftConnectionSurface, /*preferIncoming*/true, /*reverseDraw*/false);
		this._showMovePairConnections();
	},
	
	_showConnectionsFromGrid: function(fromGrid, connectionSurface, preferIncoming, reverseDraw){
		// summary:
		//		Displays connections from the given fromGrid to the middle grid
		// fromGrid:
		//		The grid that the connections are to be made from - either the left or the right grid node.
		// connectionSurface:
		//		One of the connection surfaces between grids.
		// preferIncoming:
		//		Boolean to indicate which row to prefer if there are two possible targets.
		//		If there are two possible target nodes, this indicates whether the one marked incoming
		//		is to be preferred.  'incoming' rows indicate changes that have been merged in from the
		//		source tree.
		// reverseDraw:
		//		true to indicate that the fromGrid is to the right of the middleGrid, so the connection lines
		// need to be drawn in the reverse direction.  	

		if (!this._rendered) {
			// if the _show method hasn't been called, then we have nothing to do so bail out.
			return;
		}
		
		if (!this._leftRows) {
			// the list of added or deleted rows from left hand side of the grid...
			this._leftRows = [];
			// a map of element ID -> element of things in the right hand side...
			this._rightGridIdToNodeMap = {};
			// find all the rows that have changed, in both grids.
			dojo.query(".row.blockStart", this.domNode).forEach(function(node) {
				var uri = dojo.attr(node, ROW_URI_ATTR);
				if (this._leftGrid === node.parentNode) {
					// if the node is from the left grid... 
					this._leftRows.push({
						uri: uri,
						node: node
					});
				} else {
					this._rightGridIdToNodeMap[node.id] = node;
				}
			}, this);
		}
		// now, for each row in the left hand side, find the node on the right it
		// should be connect to.
		this._leftRows.forEach(function(uriAndNode) {
			var uri = uriAndNode.uri;
			var fromBlockStartRowNode = uriAndNode.node;
			var middleIdPair = this._makeMiddleIdPair(uri);
			var incomingNode, nativeNode, targetNode;
			
			var matchingRightNodes = [],
				first = this._rightGridIdToNodeMap[middleIdPair[0]],
				second = this._rightGridIdToNodeMap[middleIdPair[1]];
				
			if (first) {
				matchingRightNodes.push(first);
			}
			
			if (second) {
				matchingRightNodes.push(second);
			}
			
			matchingRightNodes.forEach(function(candidateNode){
				if (this._isIncomingId(candidateNode.id)) {
					incomingNode = candidateNode;
				} else {
					nativeNode = candidateNode;
				}
			}, this);
			// We could have one or two matches.  If there are two matches, one of them should be
			// marked as incoming.  We favour one or other type, depending on the flag we have been given.
			if (preferIncoming) {
				targetNode = incomingNode || nativeNode;
			} else {
				targetNode = nativeNode || incomingNode;
			}
			if (targetNode) {
				// Ensure the target is a row node - if it is a placeholder entry, get the corresponding row.
				targetNode = DomUtils.findAncestorByClassInclusive(targetNode, 'row', this._middleGrid);
				this._drawGridToGridConnection(fromBlockStartRowNode, targetNode, connectionSurface, reverseDraw);
			}
		}, this);
	},
	
	_drawGridToGridConnection: function(fromNode, toNode, connectionSurface, reverseDraw) {
		
		var swapTempNode, leftEndpointY, rightEndpointY, lineColor = CONNECTION_LINE_COLOR;
		if (reverseDraw) {
			swapTempNode = fromNode;
			fromNode = toNode;
			toNode = swapTempNode;
		}
		// We now know that fromNode is to the left of toNode, so always draw left to right.
		leftEndpointY = fromNode.offsetTop - fromNode.offsetParent.scrollTop + (Math.min(CONNECTION_OFFSET, fromNode.offsetHeight/2));
		rightEndpointY = toNode.offsetTop - toNode.offsetParent.scrollTop + (Math.min(CONNECTION_OFFSET, toNode.offsetHeight/2));
		
		if (dojo.hasClass(fromNode, "added")) {
			lineColor = CONNECTION_LINE_COLOR_ADDED;
		} else if (dojo.hasClass(fromNode, "moved")) {
			lineColor = CONNECTION_LINE_COLOR_MOVED;
		}
		
		// draw a line connecting these points		
		var curve = connectionSurface.createPath();
		curve = curve.moveTo(0, leftEndpointY)
				.curveTo(CONNECTION_CURVE_CONTROL_LEFT, leftEndpointY, CONNECTION_CURVE_CONTROL_RIGHT, rightEndpointY, CONNECTION_SURFACE_WIDTH, rightEndpointY)
				.setStroke({
					color: lineColor,
					width: 1,
					style:  "Solid"
				});
		this._displayedConnections.push(curve);
		
		/*curve = connectionSurface.createPath()
					.moveTo(CONNECTION_SURFACE_WIDTH, rightEndpointY)
					.lineTo(CONNECTION_SURFACE_WIDTH-CONNECTION_ARROW_WIDTH, rightEndpointY-CONNECTION_ARROW_HEIGHT)
					.lineTo(CONNECTION_SURFACE_WIDTH-CONNECTION_ARROW_WIDTH, rightEndpointY+CONNECTION_ARROW_HEIGHT)
					.closePath()
					.setFill(lineColor);*/
		//this._displayedConnections.push(curve);
		
	},
	
	_showMovePairConnections: function() {
		// summary:
		//		Draw connections between the pairs of candidate locations for a move.
		var moveBlockPairs = {},
			node;
		dojo.query('.row.moved.blockStart', this._middleGrid).forEach(function(moveBlockRowNode){
			var uri = dojo.attr(moveBlockRowNode, ROW_URI_ATTR);
			var movePair = moveBlockPairs[uri] || (moveBlockPairs[uri] = {});
			if (this._isIncomingId(moveBlockRowNode.id)) {
				movePair.incoming = moveBlockRowNode;
			} else {
				movePair["native"] = moveBlockRowNode;
			}
		}, this);
		
		// Now we have an index containing pairs of corresponding nodes.  Each pair should have an incoming node and a native node
		// with one being a placeholder and one being the displayed block.
		for (var uri in moveBlockPairs) {
			var pair = moveBlockPairs[uri];
			this._drawMoveConnector(pair);
		}
		
		if (this._highlightedMoveConnectorEndpoints) {
			this._highlightMovePair(this._highlightedMoveConnectorEndpoints, true);
		}
		
		this._drawSelectedMoveHighlight();
	},
	
	_drawMoveConnector: function(rowPair, highlight) {
		// summary:
		//		Draw the connecting lines between a move block and the corresponding merge placeholder.
		// rowPair:
		//		An object with fields incoming and native, which are the row nodes to be linked.
		if (rowPair.incoming && rowPair["native"]) {
			var node = rowPair.incoming;
			var incomingEndpointY = node.offsetTop - node.offsetParent.scrollTop + (Math.min(CONNECTION_OFFSET, node.offsetHeight/2));
			node = rowPair["native"];
			var nativeEndpointY = node.offsetTop - node.offsetParent.scrollTop + (Math.min(CONNECTION_OFFSET, node.offsetHeight/2));
			var color = highlight ? CONNECTION_MOVE_PAIR_LINE_COLOR_BOLD : CONNECTION_MOVE_PAIR_LINE_COLOR;
			var connectionsArray = highlight ? this._moveHighlightConnections : this._displayedConnections;
			
			// draw two lines connecting these points
			// Line on left area
			var curve = this._leftConnectionSurface.createPath();
			curve = curve.moveTo(CONNECTION_SURFACE_WIDTH, incomingEndpointY)
					.curveTo(CONNECTION_SURFACE_WIDTH, incomingEndpointY, CONNECTION_SURFACE_WIDTH, incomingEndpointY, CONNECTION_ARROW_CURVE_END_RIGHT, incomingEndpointY)
					.curveTo(CONNECTION_MOVE_PAIR_CURVE_CONTROL_RIGHT, incomingEndpointY, CONNECTION_MOVE_PAIR_CURVE_CONTROL_RIGHT, nativeEndpointY, CONNECTION_ARROW_CURVE_END_RIGHT, nativeEndpointY)
					.curveTo(CONNECTION_SURFACE_WIDTH, nativeEndpointY, CONNECTION_SURFACE_WIDTH, nativeEndpointY, CONNECTION_SURFACE_WIDTH, nativeEndpointY)
					.setStroke({
						color: color,
						width: 1,
						style:  "Solid"
					});
			connectionsArray.push(curve);
			// Draw Arrow
			curve = this._leftConnectionSurface.createPath()
					.moveTo(CONNECTION_SURFACE_WIDTH, incomingEndpointY)
					.lineTo(CONNECTION_SURFACE_WIDTH-CONNECTION_ARROW_WIDTH, incomingEndpointY-CONNECTION_ARROW_HEIGHT)
					.lineTo(CONNECTION_SURFACE_WIDTH-CONNECTION_ARROW_WIDTH, incomingEndpointY+CONNECTION_ARROW_HEIGHT)
					.closePath()
					.setFill(color);
			connectionsArray.push(curve);
		}
	},
	
	_onGridClick: function(gridNode, event) {
		// summary:
		//		Handles a click within left or middle grid, which can
		//		cause various actions to happen, depending on exactly where the
		//		click occurred.
		// gridNode:
		//		The grid in which the click happened.
		// event: standard click event.
		var target = event.target,
			// find the data tree that corresponds to the gridNode
			treeRootNode = (gridNode === this._middleGrid) ? this._middleTree : this._leftTree,
			isLeftGrid = (gridNode === this._leftGrid),
			contentColumnNode,
			rowNode;
		
		dojo.stopEvent(event);
		
		// Check for clicks on show/hide links in the row
		if (dojo.hasClass(target, "showContentLink")){
			// Show the content inline
			contentColumnNode = DomUtils.findAncestorByClassInclusive(target, "contentColumn", gridNode);
			rowNode = DomUtils.findAncestorByAttr(target, ROW_URI_ATTR, gridNode);
			this._showContent(contentColumnNode, rowNode, target);
			return;
		}
		if (dojo.hasClass(target, "hideChangesLink")){
			contentColumnNode = DomUtils.findAncestorByClassInclusive(target, "contentColumn", gridNode);
			//this._hideContent(contentColumnNode);
			this._hideChanges(contentColumnNode);
			return;
		}

		
		if (this.showMoveActionButtons && dojo.hasClass(target, "moveToggleButton")){
			rowNode = DomUtils.findAncestorByAttr(target, ROW_URI_ATTR, gridNode);
			this._toggleMoveState(rowNode, treeRootNode, gridNode);
			this._showConnections();
			return;
		}
		
		// It's a click somewhere else in the grid.
		rowNode = DomUtils.findAncestorByAttr(target, ROW_URI_ATTR, gridNode);
		if (!rowNode) {
			return;
		}
		var placeholderNode = DomUtils.findAncestorByClassInclusive(target, "addedPlaceholder", rowNode)
				|| DomUtils.findAncestorByClassInclusive(target, "removedPlaceholder", rowNode);
		if (placeholderNode) {
			// click was on placeholder line between rows.  Use the placeholder that is within a row.
			rowNode = placeholderNode;
		}
		this._selectRow(isLeftGrid, rowNode);
		this._scrollOppositeSelection(isLeftGrid, rowNode);

	},
	
	_showContent: function(contentColumnNode, rowNode, showChangesLinkNode) {
		// summary:
		//		Replaces the content of the given content column node with
		//		calculated changes for the artifact on a row.
		
		// If the changes have previously been displayed, and are now hidden, show them again
		//var extendedContentNode = dojo.query(".extendedContent", contentColumnNode)[0];
		var changesContentNode = dojo.query(".changesContent", contentColumnNode)[0];
		var initialContentNode = dojo.query(".initialContent", contentColumnNode)[0];
		if (changesContentNode) {
			// no need to fetch anything - just show the nodes.
			dojo.removeClass(initialContentNode, "show");
			//dojo.addClass(extendedContentNode, "show");
			dojo.addClass(changesContentNode, "show");
			this._redrawSelectionOverlays();
			this._showConnections();
			this._renderAnnotationsIfHeightChanged();
			return;
		}
		
		// Calculate the width offset for the row.  This is used to calculate the width of the
		// new content when first added, and upon resize.
		var widthAdjust = dojo.attr(rowNode, ROW_CONTENT_WIDTH_ADJUST_ATTR);
		if (!widthAdjust) {
			var widthNode = dojo.query(".textContent", rowNode)[0];
			if (widthNode) {
				dojo.attr(rowNode, ROW_CONTENT_WIDTH_ADJUST_ATTR, rowNode.offsetWidth - widthNode.offsetWidth);
			}
		}
		
		// Call our rowContentProvider to get something to add to the content node.
		var artifactURI = dojo.attr(rowNode, ROW_URI_ATTR);
		
		var displayType, valueChanged;
		if (dojo.hasClass(rowNode, "added")) {
			displayType = CompareConstants.DISPLAY_AS_ADDED;
		} else if (dojo.hasClass(rowNode, "removed")) {
			displayType = CompareConstants.DISPLAY_AS_DELETED;
		} else if (dojo.hasClass(rowNode, "valueChanged")) {
			displayType = CompareConstants.DISPLAY_AS_CHANGED;
			valueChanged = true;
		} else {
			displayType = CompareConstants.DISPLAY_AS_NOT_CHANGED;
		}
		extendedContentNode = dojo.create('div', {"class":"displayContent extendedContent"}, contentColumnNode);
		
		// Add a link/button.  Exact function depends on whether the value in the row has changes or not.
		var selectorText = mscMessages.hideContent,
			selectorClasses = "hideContentLink";
		if (valueChanged) {
			selectorText = mscMessages.hideChanges;
			selectorClasses += " hideChanges";
		}
		var selector = dojo.create('div', {"class":"contentSelector"});
		var link = dojo.create('a', {"class":selectorClasses,"tabIndex":"-1"}, selector);
		link.setAttribute('href', '#');
		link.appendChild(document.createTextNode(selectorText));
		extendedContentNode.appendChild(selector);
		
		// get the changes and display them
		if (this._artifactContentLoader) {
			this._artifactContentLoader.loadContent(artifactURI, displayType, extendedContentNode)
				.then(function(){
					dojo.removeClass(initialContentNode, "show");
					dojo.addClass(extendedContentNode, "show");
				});
		} else {
			this._showChanges(false, contentColumnNode, rowNode, showChangesLinkNode);
		}
	},
	
	_toggleMoveState: function(rowNode, treeRootNode, gridNode) {
		// summary:
		//		Toggles the move position of a node in the merged tree and on display.
		// rowNode:
		//		The row to be toggled.
		// treeRootNode:
		//		The node in the data model tree corresponding to the clicked row.
		// gridNode:
		//		The display grid that was clicked.
		var uri = dojo.attr(rowNode, ROW_URI_ATTR);
		var treeNode = ModuleCompareUtils.getTreeNode(uri, treeRootNode);
		var toggled = ModuleCompareUtils.swapMove(treeNode);
		if (!toggled) {
			console.info("Cannot toggle move for row:", uri);
			return;
		}
		
		// Update the display to redraw the rows.  We cannot easily just move DOM nodes around, because toggling a move
		// might change the move nesting depth of many rows on the display.  So just redraw.
		// We are assuming here that this is the middle grid.  Something's gone wrong if we are doing this on a different grid.
		this._drawGrid(gridNode, treeRootNode);
		this.onMergeTreeChange();
	},
	
	_addMoveHoverControls: function(node) {
		// summary:
		//		Adds handlers to deal with hovering on a moved row.
		
		// Add the move toggle button if required.
		if (this.showMoveActionButtons) {
			dojo.create('div', {"class": "moveToggleButton"}, node);
		}
		// wait until the row has been added to the DOM before looking up the tree to find ancestors
		window.setTimeout(dojo.hitch(this, function(){
			var rowNode = DomUtils.findAncestorByClassInclusive(node, "row", this.domNode);
			this._moveToggleConnectHandles.push(this.connect(rowNode, 'onmouseenter', dojo.partial(this._onMoveToggleHover, true)));
			this._moveToggleConnectHandles.push(this.connect(rowNode, 'onmouseleave', dojo.partial(this._onMoveToggleHover, false)));
		}));
	},
	
	_onMoveToggleHover: function(highlightOn, event) {
		dojo.stopEvent(event);
		var target = event.target,
			row = DomUtils.findAncestorByClassInclusive(target, 'row', this._middleGrid),
			rowURI = dojo.attr(row, ROW_URI_ATTR),
			idPair = this._makeMiddleIdPair(rowURI);
		this._highlightMovePair(idPair, highlightOn);
	},
	
	onSelect: function(args) {
		delete this._selectedMovePair;
		this._clearMoveHighlightConnection();
		this._showMovePairConnections();
		this.inherited(arguments);
	},
	
	_highlightMovePair: function(idPair, highlightOn) {
		var moveEndpointNodes = {};
		dojo.query("#"+idPair[0]+",#"+idPair[1], this._middleGrid).forEach(function(rowNode, index){
			dojo.toggleClass(rowNode, "hover", highlightOn);
			if (index) {
				moveEndpointNodes.incoming = rowNode;
			} else {
				moveEndpointNodes["native"] = rowNode;
			}
		});
		
		
		this._clearMoveHighlightConnection();
		
		this._drawSelectedMoveHighlight();
		
		if (highlightOn) {
			this._highlightedMoveConnectorEndpoints = idPair;
			this._drawMoveConnector(moveEndpointNodes, true);
		} else {
			this._highlightedMoveConnectorEndpoints = null;
		}
	},

	_drawSelectedMoveHighlight: function() {
		if (this._rightSelection) {
			if (!this._selectedMovePair && dojo.hasClass(this._rightSelection.rowNode, "moved")) {
				this._selectedMovePair = {};
				var selectedRow = dojo.attr(this._rightSelection.rowNode, ROW_URI_ATTR);

				// find the opposite end, 
				dojo.query("[data-rdm-uri="+selectedRow+"]", this._middleGrid).forEach(function(node) {
					if (this._isIncomingId(node.id)) {
						this._selectedMovePair.incoming = node;
					} else {
						this._selectedMovePair["native"] = node;
					}
				}, this);
			}
			
			if (this._selectedMovePair) {
				this._drawMoveConnector(this._selectedMovePair, true);
			}
		}
		
	},

	_onLeftGridScroll: function() {
		this._showConnections();
	},
	
	_onMiddleGridScroll: function() {
		this._showConnections();
	},
		
	_makeLeftNodeId: function(uri) {
		return this._LEFT_ID_PREFIX + uri.substr(uri.lastIndexOf('/')+1);
	},
	
	_makeMiddleNodeId: function(uri, isIncoming) {
		return (isIncoming ? this._MIDDLE_INCOMING_ID_PREFIX : this._MIDDLE_ID_PREFIX) + uri.substr(uri.lastIndexOf('/')+1);
	},
	
	_makeMiddleIdPair: function(uri) {
		// gets a pair (in an array) of IDs that are possible IDs in the middle tree.
		var uniquePart = uri.substr(uri.lastIndexOf('/')+1);
		return [
			this._MIDDLE_ID_PREFIX + uniquePart,
			this._MIDDLE_INCOMING_ID_PREFIX + uniquePart
		];
	},
	
	_isIncomingId: function(id) {
		return (id.lastIndexOf(this._MIDDLE_INCOMING_ID_PREFIX, 0) === 0);   // more efficient way to test for match at start than string1.indexOf(string2)===0
	},
	
	_getRightIdFromLeft: function(leftNodeId, isIncoming) {
		// Summary:
		//  Called from the base class _selectRow, so ignore any linter
		//  errors, it is needed.
		if (isIncoming) {
			return leftNodeId.replace("_L_", "_MI_");
		} else {
			return leftNodeId.replace("_L_", "_M_");
		}
	},
	
	_getLeftIdFromRight: function(rightNodeId) {
		// Summary:
		//  Called from the base class _selectRow, so ignore any linter
		//  errors, it is needed.
		rightNodeId = rightNodeId.replace("_MI_", "_L_");
		return rightNodeId.replace("_M_", "_L_");
	}
});

com.ibm.rdm.web.merge.ModuleStructureMergeWidget.SECTION_NUMBER_LEFT_TREE_ARGS = {
		isHeading: function(node) {
			return node.structureNode.isHeading;
		},
		
		setSectionNumber: function(node, sectionNumber) {
			node.structureNode.sectionNumberPreMerge = sectionNumber;
		},
		
		_accumulateChildrenAndExcludedDescendants: function(nodes, children) {
			var node, i, l;
			if (Array.isArray(nodes)) {
				for (i=0, l=nodes.length; i<l; i++) {
					node = nodes[i];
					if (node.excluded || !node.content) {
						this._accumulateChildrenAndExcludedDescendants(node.children, children);
					} else {
						children.push(node);
					}
				}
			} else if (nodes) {
				this._accumulateChildrenAndExcludedDescendants(nodes.children, children);
			}
		},
				
		getChildren: function(node) {
			var children = [];
			this._accumulateChildrenAndExcludedDescendants(node, children);
			return children;			
		}
}; 
	
	
	com.ibm.rdm.web.merge.ModuleStructureMergeWidget.SECTION_NUMBER_MERGE_TREE_ARGS = dojo.delegate(com.ibm.rdm.web.merge.ModuleStructureMergeWidget.SECTION_NUMBER_LEFT_TREE_ARGS, {
		setSectionNumber: function(node, sectionNumber) {		
			node.structureNode.section = sectionNumber;
		}
	});
	


})();