/*
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2009, 2016.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 */
(function() {
dojo.provide("com.ibm.rdm.web.review.ParticipantResultSaver");
dojo.require("dojo.date");
dojo.require("dojo.date.locale");

dojo.require("com.ibm.rdm.web.common.ErrorHandler");
dojo.require("com.ibm.rdm.web.common.MessageHandler");
dojo.require("com.ibm.rdm.web.common.MimeTypes");
dojo.require("com.ibm.rdm.web.review.reviewTypes");
dojo.require("com.ibm.rdm.web.server.Headers");
dojo.require("com.ibm.rdm.web.server.XhrUtils");
dojo.require("com.ibm.rdm.web.server.client.RDMClient");
dojo.require("com.ibm.rdm.web.util.queryutils");

dojo.requireLocalization("com.ibm.rdm.web.review", "reviewAndApprovalCommon");


var ErrorHandler = com.ibm.rdm.web.common.ErrorHandler;
var Headers = com.ibm.rdm.web.server.Headers;
var MimeTypes = com.ibm.rdm.web.common.MimeTypes;
var RDMClient = com.ibm.rdm.web.server.client.RDMClient;
var XhrUtils = com.ibm.rdm.web.server.XhrUtils;
var messages = dojo.i18n.getLocalization("com.ibm.rdm.web.review", "reviewAndApprovalCommon");
var queryutils = com.ibm.rdm.web.util.queryutils;
var reviewTypes = com.ibm.rdm.web.review.reviewTypes;


// This class is responsible for saving away the ParticipantResult artifact for
// the current user.  Here's an example of the artifact that gets saved.
//
// <?xml version="1.0" encoding="UTF-8"?>
// <rdf:RDF
//    xmlns:rrmReview="http://www.ibm.com/xmlns/rrm/reviews/1.0/"
//    xmlns:dc="http://purl.org/dc/elements/1.1/"
//    xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
//    <rrmReview:ParticipantResult rdf:about="https://localhost:9443/rdm/reviews/results/_VQ5n8WgLEd6NlcEl9xCf0g">
//      <rrmReview:associatedReview rdf:resource="https://localhost:9443/rdm/reviews/_LvKL8GgLEd6NlcEl9xCf0g" />
//      <rrmReview:hasParticipant rdf:resource="https://localhost:9443/jazz/users/Anne" />
//      <rrmReview:reviewRole>Approver</rrmReview:reviewRole>
//      <rrmReview:isDone>false</rrmReview:isDone>
//      <rrmReview:hasArtifactResult>
//        <rrmReview:ArtifactResult>
//          <rrmReview:artifactId>https://localhost:9443/rdm/resources/_vFKIYFxTEd6WMqY891WRNA?revision=_vG8RElxTEd6WMqY891WRNA</rrmReview:artifactId>
//          <rrmReview:participantResultStatus>Abstained</rrmReview:participantResultStatus>
//        </rrmReview:ArtifactResult>
//      </rrmReview:hasArtifactResult>
//      <rrmReview:hasArtifactResult>
//        <rrmReview:ArtifactResult>
//          <rrmReview:artifactId>https://localhost:9443/rdm/resources/_vT5j0WGJEd6napvdB5J99g?revision=_CJf9gWTPEd6FApb3gAdFRg</rrmReview:artifactId>
//          <rrmReview:participantResultStatus>Approved</rrmReview:participantResultStatus>
//        </rrmReview:ArtifactResult>
//      </rrmReview:hasArtifactResult>
//    </rrmReview:ParticipantResult>
// </rdf:RDF>

// Representation used on 9/24/09 for a started formal review:
// <rdf:RDF
//    xmlns:rrmReview="http://www.ibm.com/xmlns/rrm/reviews/1.0/"
//    xmlns:dc="http://purl.org/dc/elements/1.1/"
//    xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
//    <rrmReview:ParticipantResult rdf:about="https://localhost:9443/rdm/reviews/results/_x2NIoKk_Ed6MXPez46OKWA">
//      <rrmReview:associatedReview rdf:resource="_f8ow8Kk9Ed6MXPez46OKWA" />
//      <rrmReview:hasParticipant rdf:resource="../../jazz/users/TestJazzAdmin1" />
//      <rrmReview:reviewRole>Required Reviewer</rrmReview:reviewRole>
//      <rrmReview:isDone>false</rrmReview:isDone>
//      <rrmReview:hasArtifactResult>
//        <rrmReview:ArtifactResult>
//          <rrmReview:artifactId>../resources/rrc89?revision=_U4XyQak9Ed6MXPez46OKWA</rrmReview:artifactId>
//          <rrmReview:participantResultStatus>Reviewed</rrmReview:participantResultStatus>
//        </rrmReview:ArtifactResult>
//      </rrmReview:hasArtifactResult>
//    </rrmReview:ParticipantResult>
// </rdf:RDF>


dojo.declare(
	"com.ibm.rdm.web.review.ParticipantResultSaver",
	null,
{
	// summary:
	//		This class is responsible for persisting the "participant result"
	//		artifacts that are used to represent vote casts for the different
	//		participants within a review.
	// description:
	//		Every review that's not in the Draft state will contain at least
	//		one participant.  Each of these participants uses a single
	//		participant result artifact to record his/her votes.  These
	//		participant result artifacts are created "on demand"; it's
	//		only after a participant casts his/her first vote that a new
	//		participant result artifact will be POST'd for this user in this
	//		review.  Subsequent votes cast by this participant will just update
	//		his/her existing participant result artifact via PUT calls.

	///////////////////////////////////////////////////////////////////////////
	// Private fields

	_reviewParser: null,

	///////////////////////////////////////////////////////////////////////////
	// API

	constructor: function(keywordArgs) {
		// summary:
		//		Parses the content found within the RDF that represents a
		//		ParticipantResult artifact.
		// keywordArgs.doc:
		//		The response returned by a doGet for the participant results
		//		associated with a review.
		this._reviewParser = keywordArgs.reviewParser;
	},

	castArtifactVotes: function(
		myChangedArtifactInfos,
		onSaveComplete,
		/*Callback?*/onSaveError,
		/*Boolean?*/skipErrors
	) {
		// summary:
		//		Save away new artifact vote casts for the the current user.
		// myChangedArtifactInfos: Array of artifactInfos
		//		Please note that the artifactInfos found in this array represent
		//		ONLY those artifactInfo's whose vote setting has changed (i.e.
		//		it is NOT intended to represent the complete list of new values
		//		to record [changed vote values as well as original vote values]).
		var createResultInfos = function(userURI, changedArtifactInfos) {
			return (
				dojo.map(
					changedArtifactInfos,
					function(artifactInfo) {
						return {
							artifact: artifactInfo.id,
							artifactStatusType: artifactInfo.myArtifactStatusType,
							userURI: userURI
						};
					}
				)
			);
		};
		var myUserURI = com.ibm.rdm.web.AUTHENTICATED_USER.uri;
		var myChangedArtifactResultInfos =
			createResultInfos(myUserURI, myChangedArtifactInfos);
		var participantResultContent =
			this._composeParticipantResult(
				myUserURI, false, null, myChangedArtifactResultInfos
			);
		var etag = this._reviewParser.getParticipantResultETag();
		this._performSaveParticipantResult(
			 myUserURI,
			 etag,
			 participantResultContent,
			onSaveComplete,
			onSaveError,
			skipErrors
		);
	},

	markMyReviewStatusAsDone: function(onSaveComplete, /*Callback?*/onSaveError) {
		// summary:
		//		Mark the current user's reviewing state as "Done" for the
		//		entire review.
		// description:
		//		Create artifactResults elements using Abstained state for all
		//		artifacts that the user has not yet started.  Also for all
		//		artifactResults associated with the user that are InProgress,
		//		move these to Abstained.  Finally add an isDone="true" attribute
		//		to the current user's participant results element and save this
		//		new version of the Review.
		//		Calls onSaveComplete when save has completed.
		var myUserURI = com.ibm.rdm.web.AUTHENTICATED_USER.uri;
		var artifactStatusType = reviewTypes.ARTIFACT_STATUS_TYPES.Abstained;
		var participantResultContent =
			this._composeParticipantResult(
				myUserURI, true, artifactStatusType, null
			);
		var etag = this._reviewParser.getParticipantResultETag();
		this._performSaveParticipantResult(
			 myUserURI,
			 etag,
			 participantResultContent,
			 onSaveComplete, onSaveError);
	},

	resetArtifactVotes: function(participantInfo,
								 isDone,
								 artifactStatusType,
								 changedArtifactResultInfos,
								 onSaveComplete,
								 /*Callback?*/onSaveError) {
		// summary:
		//		Save away new artifact vote casts for a particular user.
		// description:
		//		For the user found in 'participantInfo' save away the new
		//		vote casts that are represented by 'changedArtifactResultInfos'.
		// participantInfo:
		//		The new vote casts are to be made for the user found in this
		//		participantInfo.
		// isDone:
		//		A boolean representing whether the participant result to be
		//		updated for participant 'userURI' should record this
		//		participant as being done or not yet done.
		// artifactStatusType:
		//		When isDone is passed as true, this parameter holds the
		//		artifact status type that should be recorded for all remaining
		//		artifacts (i.e. those artifacts that this participant has not
		//		yet voted on).
		// changedArtifactResultInfos:
		//		Please note that the artifactInfos found in this array represent
		//		ONLY those artifactInfo's whose vote setting has changed (i.e.
		//		it is NOT intended to represent the complete list of new values
		//		to record [changed vote values as well as original vote values]).
		var userURI = participantInfo.userURI;
		var participantResultContent =
			this._composeParticipantResult(
				userURI, isDone, artifactStatusType, changedArtifactResultInfos
			);
		// Note by SPC: In the following we pass null as the etag (meaning
		// "*" will be used for If-Match).  This shouldn't be a problem, as the
		// review is in a Paused state just before this function is called (and
		// participants aren't allowed to cast votes when the review is in the
		// Paused state).
		// One could argue that we should pass the appropriate etag regardless ...
		// but it's expensive to do the HEAD requests to obtain this info (and
		// it appears to not really be needed for this case).
		this._performSaveParticipantResult(
			 userURI,
			 null,				// etag
			 participantResultContent,
			 onSaveComplete, onSaveError);
	},
	
	getSaveArtifactVotesArgs: function(participantInfo, isDone,
		artifactStatusType, changedArtifactResultInfos) {
		// summary:
		//		Save away new artifact vote casts for a particular user.
		// description:
		//		For the user found in 'participantInfo' save away the new
		//		vote casts that are represented by 'changedArtifactResultInfos'.
		// participantInfo:
		//		The new vote casts are to be made for the user found in this
		//		participantInfo.
		// isDone:
		//		A boolean representing whether the participant result to be
		//		updated for participant 'userURI' should record this
		//		participant as being done or not yet done.
		// artifactStatusType:
		//		When isDone is passed as true, this parameter holds the
		//		artifact status type that should be recorded for all remaining
		//		artifacts (i.e. those artifacts that this participant has not
		//		yet voted on).
		// changedArtifactResultInfos:
		//		Please note that the artifactInfos found in this array represent
		//		ONLY those artifactInfo's whose vote setting has changed (i.e.
		//		it is NOT intended to represent the complete list of new values
		//		to record [changed vote values as well as original vote values]).
		var userURI = participantInfo.userURI;
		var requestArguments = {
			xhrArgs: {},
			componentURI: this._reviewParser.getComponentUri()
		};
		var headers = {};
		var participantResultContent =
			this._composeParticipantResult(userURI, isDone, artifactStatusType,
				changedArtifactResultInfos);
		var participantResultResourceURI =
			this._reviewParser.getParticipantResultUri(userURI);

		// Set method and content
		if(participantResultResourceURI) {
			// Exists, PUT and update
			requestArguments.method = RDMClient.HTTP_METHOD_PUT;
			requestArguments.uri =
				this._reviewParser.getParticipantResultUri(userURI);
			requestArguments.putData = participantResultContent;
		}
		else {
			// Doesn't exist, POST a new one
			requestArguments.method = RDMClient.HTTP_METHOD_POST;
			requestArguments.uri =
				com.ibm.rdm.web.FRONTING_SERVICE_REVIEW_AND_APPROVAL +
				"/results";
			requestArguments.postData = participantResultContent;
		}

		// Set Headers
		headers[Headers.CONTENT_TYPE] =
			MimeTypes.CONTENT_TYPE_REVIEW_RESULT;
		headers[Headers.IF_MATCH] = "*"; // Will we ever need an etag here?

		requestArguments.xhrArgs.headers = headers;

		return requestArguments;
	 },

	///////////////////////////////////////////////////////////////////////////
	// Helpers

	_performSaveParticipantResult: function(
		userURI,
		etag,
		participantResultContent,
		onSaveComplete,
		/*Callback?*/onSaveError,
		/*Boolean?*/skipErrors
	) {
		var participantResultResourceURI = this._reviewParser.getParticipantResultUri(userURI);
		var self = this;
		var performSave = function(headerValueForIfMatch) {
			var deferred = self._saveParticipantResult(
				userURI,
				participantResultResourceURI,
				participantResultContent,
				headerValueForIfMatch,
				skipErrors === undefined || skipErrors === false ? false : true
			);
			deferred.addCallback(function(response) {
				onSaveComplete();
				return response;
			});
			deferred.addErrback(function(response) {
				if (onSaveError) {
					onSaveError(response);
				}
				return response;
			});
		};
		if (participantResultResourceURI) {
			if (etag) {
				performSave('"' + etag + '"');
			} else {
				// Note: We end up in this 'else' clause when performing the
				// "Change Status" management function.
				performSave("*");
			}
		}
		else {
			// It must be the case that this user has never previously voted
			// within the context of this review.  So we'll end up doing a POST.
			performSave("*");
		}
	},

	_saveParticipantResult: function(
		userURI,
		participantResultResourceURI,
		participantResultContent,
		headerValueForIfMatch,
		/*Boolean*/skipErrors
	) {
		// summary:
		//		Attempt to save a new version of the review with the new review
		//		content found in the passed participantResultContent parameter.
		// returns:
		//		A deferred.
		var headers = {};
		headers[Headers.OWNING_CONTEXT] = this._reviewParser.getComponentUri();
		headers[Headers.CONTENT_TYPE] = MimeTypes.CONTENT_TYPE_REVIEW_RESULT;
		headers[Headers.IF_MATCH] = headerValueForIfMatch;
		var deferred = null;
		// Do either a POST (in the case of never having voted before) or a PUT
		// (if the user has previously cast votes in the context of this
		// review).
		com.ibm.rdm.web.review.reviewUpdateMediator.reviewSaveInProgress(
			this._reviewParser);
		if (participantResultResourceURI) {
			deferred = RDMClient.doPut({
				uri: participantResultResourceURI,
				putData: participantResultContent,
				delayCallbackUntilDataIndexed: false,
				skipErrors: skipErrors,
				xhrArgs: {
					headers: headers
				}
			});
		} else {
			deferred = RDMClient.doPost({
				uri: com.ibm.rdm.web.FRONTING_SERVICE_REVIEW_AND_APPROVAL + "/results",
				postData: participantResultContent,
				delayCallbackUntilDataIndexed: false,
				skipErrors: true,
				xhrArgs: {
					headers: headers
				}
			});
		}
		var self = this;
		deferred.addCallback(this, function(response){
			if (userURI === com.ibm.rdm.web.AUTHENTICATED_USER.uri) {
				var etag = XhrUtils.getResponseEtag(response);
				if (! participantResultResourceURI) {
					// Note: When performing a POST the returned "Location" response
					// header contains the new resource URI.
					participantResultResourceURI = XhrUtils.getHeader(response, "Location");
				}
				com.ibm.rdm.web.review.reviewUpdateMediator.reviewSaveHasCompleted(
					self._reviewParser, etag, participantResultResourceURI);
			}
			return response.doc;
		});
		deferred.addErrback(this, function(response){
			var putOrPostStatus = response.ioArgs.xhr.status;
			com.ibm.rdm.web.review.reviewUpdateMediator.reviewSaveHasFailed(
				self._reviewParser, putOrPostStatus);
			ErrorHandler.warn(
				"Received abnormal response as a result of attempting to save ParticipantResult: "
				+ putOrPostStatus
			);
			// Check if this abnormal response was returned due to the fact
			// that the underlying Review has been deleted.
			var headers = [];
			headers[Headers.OWNING_CONTEXT] = this._reviewParser.getComponentUri();
			var headDeferred = RDMClient.doHead({
				uri: self._reviewParser.getSelfUri(),
				xhrArgs: { headers: headers },
				skipCache: true,
				skipErrors: true
			});
			headDeferred.addCallback(this, function(response){
				if (!skipErrors) {
					self._showCouldNotSaveVotesError(putOrPostStatus, "unableToSaveParticipantResultDetails");
				}
				return response;
			});
			headDeferred.addErrback(this, function(response){
				var status = response.ioArgs.xhr.status;
				if (status === 410) { // Note: 410 is "Gone"
					ErrorHandler.warn(
						"The underlying review '"
						+ self._reviewParser.getReviewName()
						+ "' has been deleted"
					);
					self._showReviewHasBeenDeletedError();
				} else if (status === 409) { // Note: 409 is "Conflict"
					self._showCouldNotSaveVotesError(putOrPostStatus, "unableToSaveParticipantResultDetails409");
				} else {
					self._showCouldNotSaveVotesError(putOrPostStatus, "unableToSaveParticipantResultDetails");
				}
				return response;
			});
			return response;
		});
		return deferred;
	},

	_showCouldNotSaveVotesError: function(statusCode, messageName) {
		var details = dojo.string.substitute(
			messages[messageName],
			[
				"ParticipantResultSaver._saveParticipantResult",
				this._reviewParser.getReviewName(),
				statusCode
			]
		);
		com.ibm.rdm.web.common.MessageHandler.showError(7302, null, details); //unableToSaveParticipantResult
	},

	_showReviewHasBeenDeletedError: function() {
		var details = dojo.string.substitute(
			messages["saveParticipantResultsFailedDueToDeletedReviewDetails"],
			[
				"ParticipantResultSaver._saveParticipantResult",
				this._reviewParser.getReviewName()
			]
		);
		com.ibm.rdm.web.common.MessageHandler.showError(7303, null, details); //saveParticipantResultsFailedDueToDeletedReview
	},

	_composeParticipantResult: function(
			userURI, isDone, artifactStatusType, changedArtifactResultInfos) {
		// summary:
		//		Create the XML that reflects this.reviewParser content
		//		overlayed with either artifactStatusType or
		//		changedArtifactResultInfos (whichever is passed).
		// description:
		//		Note: It's expected that only one of the last two params,
		//		artifactStatusType or changedArtifactResultInfos, will hold
		//		a non-null value.
		var userRole = this._reviewParser.getRole(userURI);
		var participantResultResourceURI =
			this._reviewParser.getParticipantResultUri(userURI);
		var about = "";
		var reviewURI = this._reviewParser.getSelfUri();
		var hasArtifactResultElementContent =
			this._composeHasArtifactResultElementContent(userURI, artifactStatusType, changedArtifactResultInfos);
		return (
			"<" + queryutils.RDF_QUALIFIER + ":RDF\r\n"
			+ "    xmlns:" + queryutils.REVIEW_QUALIFIER + "=\"" + queryutils.REVIEW_NAMESPACE + "\"\r\n"
			+ "    xmlns:" + queryutils.DC_QUALIFIER + "=\"" + queryutils.DC_NAMESPACE + "\"\r\n"
			+ "    xmlns:" + queryutils.RDF_QUALIFIER + "=\"" + queryutils.RDF_NAMESPACE + "\">\r\n"
			+ "  <" + queryutils.REVIEW_QUALIFIER + ":ParticipantResult " + queryutils.RDF_QUALIFIER + ":about=\"" + about + "\">\r\n"
			+ "    <" + queryutils.REVIEW_QUALIFIER + ":associatedReview rdf:resource=\"" + reviewURI + "\"/>\r\n"
			+ "    <" + queryutils.REVIEW_QUALIFIER + ":hasParticipant " + queryutils.RDF_QUALIFIER + ":resource=\"" + userURI + "\"/>\r\n"
			+ "    <" + queryutils.REVIEW_QUALIFIER + ":reviewRole>" + userRole + "</" + queryutils.REVIEW_QUALIFIER + ":reviewRole>\r\n"
			+ "    <" + queryutils.REVIEW_QUALIFIER + ":isDone>" + isDone + "</" + queryutils.REVIEW_QUALIFIER + ":isDone>\r\n"
			+ hasArtifactResultElementContent
			+ "  </" + queryutils.REVIEW_QUALIFIER + ":ParticipantResult>\r\n"
			+ "</" + queryutils.RDF_QUALIFIER + ":RDF>\r\n"
		);
	},

	_composeHasArtifactResultElementContent: function(
			userURI, artifactStatusType, changedArtifactResultInfos) {
		// summary:
		//		Compose the "participantResults" element for the current user.
		// description:
		//		Note: It's expected that only one of the last two params,
		//		artifactStatusType or changedArtifactResultInfos, will hold a
		//		non-null value.
		// userURI:
		//		A userURI representing the user for which we're attempting to
		//		save a participant result.
		// artifactStatusType:
		//		The "vote" that should be recorded for the remaining artifacts
		//		that have not yet been voted on.  The value is expected to be
		//		one of the values found in reviewTypes.ARTIFACT_STATUS_TYPES.
		// changedArtifactResultInfos:
		//		An array of artifactInfo property lists containing the intended
		//		artifact votes that the should be recorded for the passed user
		//		in the context of this review.
		var artifactResultInfos = this._reviewParser.getArtifactResultInfos();
		var buf = "";
		// Capture the new values that should be recorded for the passed user
		// as singleUserArtifactResultInfos.
		var singleUserArtifactResultInfos = (
			artifactStatusType ?
			this._composeArtifactResultInfosByDefaultingStatusForRemainingArtifacts(
				userURI, artifactStatusType) :
			this._composeArtifactResultInfosUsingChangedArtifactResultInfos(
				userURI, changedArtifactResultInfos)
		);
		// Fill in buf.
		dojo.forEach(
			singleUserArtifactResultInfos,
			function(userArtifactResultInfo) {
				buf += this._composeArtifactResultElementContent(
					userArtifactResultInfo);
			},
			this
		);
		return buf;
	},

	_composeArtifactResultElementContent: function(artifactResultInfo) {
		var artifactId = artifactResultInfo.artifact;
		var artifactStatusType = artifactResultInfo.artifactStatusType;
		return (
			"    <" + queryutils.REVIEW_QUALIFIER + ":hasArtifactResult>\r\n"
			+ "      <" + queryutils.REVIEW_QUALIFIER + ":ArtifactResult>\r\n"
			+ "        <" + queryutils.REVIEW_QUALIFIER + ":artifactId>" + artifactId
			+ "</" + queryutils.REVIEW_QUALIFIER + ":artifactId>\r\n"
			+ "        <" + queryutils.REVIEW_QUALIFIER + ":participantResultStatus>" + artifactStatusType
			+ "</" + queryutils.REVIEW_QUALIFIER + ":participantResultStatus>\r\n"
			+ "      </" + queryutils.REVIEW_QUALIFIER + ":ArtifactResult>\r\n"
			+ "    </" + queryutils.REVIEW_QUALIFIER + ":hasArtifactResult>\r\n"
		);
	},

	_composeArtifactResultInfosUsingChangedArtifactResultInfos: function(
			userURI, changedArtifactResultInfos) {
		// summary:
		//		Get the artifactResultInfos that should be recorded for the
		//		passed user if he/she were to make the vote casting changes
		//		found in changedArtifactResultInfos.
		var newArtifactResultInfos = [];
		var artifactInfos = this._reviewParser.getArtifactInfos();
		// A map for the previously recorded result infos for this user.
		var mapSingleUserArtifactIdToArtifactResultInfo =
			com.ibm.rdm.web.review.reviewutils.createSingleUserMapOfArtifactIdToArtifactResultInfo(
				this._reviewParser, userURI);
		// A map for the newly changed result infos for this user.
		var mapArtifactIdToArtifactResultInfo = [];
		// Fill in mapArtifactIdToArtifactResultInfo.
		dojo.forEach(
			changedArtifactResultInfos,
			function(changedResultInfo) {
				mapArtifactIdToArtifactResultInfo[changedResultInfo.artifact] =
					changedResultInfo;
			}
		);
		// Fill in newArtifactResultInfos.
		dojo.forEach(
			artifactInfos,
			function(artifactInfo) {
				// Set artifactResultInfoToUse to the passed artifactResultInfo
				// found in the passed changedArtifactResultInfos, or if not found in
				// this array, set it instead to the last recorded value (i.e.
				// the value found in this._reviewParser).
				var artifactResultInfoToUse = (
					mapArtifactIdToArtifactResultInfo[artifactInfo.id]
					|| mapSingleUserArtifactIdToArtifactResultInfo[artifactInfo.id]
				);
				// Note: Please note that we ONLY record a NotStarted vote if
				// it was either originally recorded this way, or it's
				// explicitly specified as NotStarted in the
				// changedArtifactResultInfos.
				if (artifactResultInfoToUse) {
					newArtifactResultInfos.push({
						userId: userURI,
						artifact: artifactResultInfoToUse.artifact,
						artifactStatusType: artifactResultInfoToUse.artifactStatusType
					});
				}
			}
		);
		return newArtifactResultInfos;
	},

	_composeArtifactResultInfosByDefaultingStatusForRemainingArtifacts: function(
			userURI, artifactStatusType) {
		// summary:
		//		Get the artifactResultInfos that should be recorded for a
		//		particular user by defaulting all remaining artifacts
		//		statuses to the passed 'artifactStatusType'.
		// example:
		//		Consider a review containing three artifacts in which the user
		//		has only cast a vote for one of these artifacts.  As a result
		//		of calling this function, an array containing three
		//		artifactResultInfos would be returned; one for the existing
		//		artifact vote previously cast and two for "default" votes for
		//		the remaining two artifacts that have not yet had a vote cast.
		var newArtifactResultInfos = [];
		var artifactInfos = this._reviewParser.getArtifactInfos();
		var mapSingleUserArtifactIdToArtifactResultInfo =
			com.ibm.rdm.web.review.reviewutils.createSingleUserMapOfArtifactIdToArtifactResultInfo(
				this._reviewParser, userURI);
		dojo.forEach(
			artifactInfos,
			function(artifactInfo) {
				var artifactResultInfo = mapSingleUserArtifactIdToArtifactResultInfo[artifactInfo.id];
				var artifactStatusToUse = (
					(!artifactResultInfo
						|| artifactResultInfo.artifactStatusType === reviewTypes.ARTIFACT_STATUS_TYPES.NotStarted
						|| artifactResultInfo.artifactStatusType === reviewTypes.ARTIFACT_STATUS_TYPES.InProgress) ?
					artifactStatusType :
					artifactResultInfo.artifactStatusType
				);
				newArtifactResultInfos.push({
					userId: userURI,
					artifact: artifactInfo.id,
					artifactStatusType: artifactStatusToUse
				});
			}
		);
		return newArtifactResultInfos;
	}
});


})();
